﻿using System;
using System.IO;
using System.Reflection;

namespace Scrabble
{
    class BitStream
    {
        private object          _locker = new object();
        private byte            _offset;
        private long            _pos_bit;
        private int             _pos_byte;
        private long            _size_bit;
        private int             _size_byte;
        private MemoryStream    _stream;
        private BinaryReader    _reader;
        private BinaryWriter    _writer;

        static byte[] ConvertToAlpha    = { 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31,
                                            31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31,
                                            31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31,
                                            31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31,
                                            31, 00, 01, 02, 03, 04, 05, 06, 07, 08, 09, 10, 11, 12, 13, 14,
                                            15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 31, 31, 31, 31, 31,
                                            31, 00, 01, 02, 03, 04, 05, 06, 07, 08, 09, 10, 11, 12, 13, 14,
                                            15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 31, 31, 31, 31, 31 };

        static byte[] ConvertFromAlpha  = { 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80,
                                            81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 32, 32, 32, 32, 32, 32 };

        public BitStream()
        {
            this._offset    = 0;
            this._pos_bit   = 0;
            this._pos_byte  = 0;
            this._size_bit  = 0;
            this._size_byte = 0;
            this._stream    = new MemoryStream();
            this._reader    = new BinaryReader(this._stream);
            this._writer    = new BinaryWriter(this._stream);
        }

        public void Dispose()
        {
            this._offset    = 0;
            this._pos_bit   = 0;
            this._pos_byte  = 0;
            this._size_bit  = 0;
            this._size_byte = 0;

            this._reader.Close();
            this._writer.Close();
            this._stream.Close();
        }

        public void WriteBit(int bit)
        {
            byte value;

            if ((this._size_bit & 7) == 0)
            {
                value = (byte)(bit << 7);
            }
            else
            {
                this._stream.Seek(this._pos_byte, SeekOrigin.Begin);

                value  = this._reader.ReadByte();
                value ^= (byte)(value & (1 << (7 - this._offset)));
                value |= (byte)(bit << (7 - this._offset));
            }

            this._stream.Seek(this._pos_byte, SeekOrigin.Begin);
            this._writer.Write(value);
            this._pos_bit++;
            this._size_bit  = (this._pos_bit > this._size_bit ? this._pos_bit : this._size_bit);
            this._size_byte = (int)this._stream.Length;  //(this._size_bit >> 3) + ((this._size_bit & 7) > 0 ? 1 : 0);
            this._pos_byte  = (int)(this._pos_bit >> 3);
            this._offset    = (byte)(this._pos_bit & 7);
        }
        public void Write(bool       data                              )
        {
            lock (this._locker)
            {
                this.WriteBit(data ? 1 : 0);
            }
        }
        public void Write(byte       data, int bits                    )
        {
            if ((bits < 1) || (bits > 8))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((data & ((byte)1 << bit)) >> bit);
                }
            }
        }
        public void Write(sbyte      data, int bits                    )
        {
            if ((bits < 1) || (bits > 8))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)(((byte)data & ((byte)1 << bit)) >> bit));
                }
            }
        }
        public void Write(short      data, int bits                    )
        {
            if ((bits < 1) || (bits > 16))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)(((ushort)data & ((ushort)1 << bit)) >> bit));
                }
            }
        }
        public void Write(ushort     data, int bits                    )
        {
            if ((bits < 1) || (bits > 16))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)((data & ((ushort)1 << bit)) >> bit));
                }
            }
        }
        public void Write(int        data, int bits                    )
        {
            if ((bits < 1) || (bits > 32))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)(((uint)data & ((uint)1 << bit)) >> bit));
                }
            }
        }
        public void Write(uint       data, int bits                    )
        {
            if ((bits < 1) || (bits > 32))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)((data & ((uint)1 << bit)) >> bit));
                }
            }
        }
        public void Write(long       data, int bits                    )
        {
            if ((bits < 1) || (bits > 64))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)(((ulong)data & ((ulong)1 << bit)) >> bit));
                }
            }
        }
        public void Write(ulong      data, int bits                    )
        {
            if ((bits < 1) || (bits > 64))
            {
                return;
            }

            lock (this._locker)
            {
                for (int bit = bits - 1; bit >= 0; bit--)
                {
                    this.WriteBit((int)((data & ((ulong)1 << bit)) >> bit));
                }
            }
        }
        public void Write(DateTime   data                              )
        {
            this.Write((long)data.ToBinary(), 64);
        }
        public void Write(char       data, int bits                    )
        {
            if ((bits < 5) || (bits > 32))
            {
                return;
            }

            lock (this._locker)
            {
                int b = (int)data;

                for (int x = bits - 1; x >= 0; x--)
                {
                    this.WriteBit((b & (1 << x)) >> x);
                }
            }
        }
        public void Write(string     data, int bits)
        {
            if ((bits < 5) || (bits > 6))
            {
                return;
            }

            lock (this._locker)
            {
                for (int x = 0; x < data.Length; x++)
                {
                    byte b = ConvertToAlpha[(byte)data[x]];

                    for (int y = bits - 1; y >= 0; y--)
                    {
                        this.WriteBit((b & (1 << y)) >> y);
                    }
                }
            }
        }
        public void Write(bool    [] data                              )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x]);
            }
        }
        public void Write(byte    [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(sbyte   [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(short   [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(ushort  [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(int     [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(uint    [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(long    [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(ulong   [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(DateTime[] data                              )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x]);
            }
        }
        public void Write(char    [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }
        public void Write(string  [] data, int bits                    )
        {
            for (int x = 0; x < data.Length; x++)
            {
                this.Write(data[x], bits);
            }
        }

        public int      ReadBit      ()
        {
            this._stream.Seek(this._pos_byte, SeekOrigin.Begin);

            byte value      = this._reader.ReadByte();
            byte bit        = (byte)(value & (1 << (7 - this._offset)));
            int  result     = ((value & (1 << (7 - this._offset))) >> (7 - this._offset));

            lock (this._locker)
            {
                this._pos_bit++;
                this._pos_byte = (int)(this._pos_bit >> 3);
                this._offset   = (byte)(this._pos_bit & 7);
            }
            
            return result;
        }
        public bool     ReadBoolean  ()
        {
            return (this.ReadBit() == 1);
        }
        public object   ReadInteger  (bool signed, int bits)
        {
            if ((bits < 1) || (bits > 64))
            {
                return 0;
            }

            ulong value = 0;

            for (int x = bits - 1; x >= 0; x--)
            {
                ulong bit = (ulong)this.ReadBit();

                value |= bit << x;

                if (signed && (x == (bits - 1)))
                {
                    for (int y = x + 1; y < 64; y++)
                    {
                        value |= bit << y;
                    }
                }
            }

            if (signed)
            {
                if (bits <= 8)
                {
                    return (sbyte)(value & 0xFF);
                }
                else if (bits <= 16)
                {
                    return (short)(value & 0xFFFF);
                }
                else if (bits <= 32)
                {
                    return (int)(value & 0xFFFFFFFF);
                }
                else
                {
                    return (long)value;
                }
            }
            else
            {
                if (bits <= 8)
                {
                    return (byte)(value & 0xFF);
                }
                else if (bits <= 16)
                {
                    return (ushort)(value & 0xFFFF);
                }
                else if (bits <= 32)
                {
                    return (uint)(value & 0xFFFFFFFF);
                }
                else
                {
                    return value;
                }
            }
        }
        public DateTime ReadDateTime ()
        {
            return DateTime.FromBinary((long)this.ReadInteger(true, 64));
        }
        public char     ReadChar     (int bits)
        {
            if ((bits < 7) || (bits > 32))
            {
                return '\0';
            }

            return (char)this.ReadInteger(false, bits);
        }
        public string   ReadString   (int length, int bits)
        {
            if ((length < 0) || (bits < 5) || (bits > 6))
            {
                return "";
            }

            string result = "";

            for (int x = 0; x < length; x++)
            {
                result += new string(new char[] { (char)ConvertFromAlpha[(byte)this.ReadInteger(false, bits)] });
            }

            return result;
        }

        public bool     AtEndOfStream
        {
            get
            {
                return (_pos_byte == (_size_byte - 1));
            }
        }
        public void     Seek         (long Offset, SeekOrigin Origin)
        {
            lock (this._locker)
            {
                switch (Origin)
                {
                    case SeekOrigin.Begin:   this._pos_bit  = Offset; break;
                    case SeekOrigin.Current: this._pos_bit += Offset; this._pos_bit %= this._size_bit; break;
                    case SeekOrigin.End:     this._pos_bit  = this._size_bit - Offset; break;
                }

                this._pos_bit += this._size_bit;
                this._pos_bit %= this._size_bit;
                this._pos_byte =  (int)(this._pos_bit >> 3);
                this._offset   = (byte)(this._pos_bit &  7);
            }
        }
        public byte[]   GetBytes     ()
        {
            this._stream.Seek(0, SeekOrigin.Begin);
            return this._reader.ReadBytes((int)this._stream.Length);
        }
        public void     LoadFromFile ()
        {
            this.Dispose();

            this._stream    = new MemoryStream();
            this._reader    = new BinaryReader(this._stream);
            this._writer    = new BinaryWriter(this._stream);
            Stream file     = Assembly.GetExecutingAssembly().GetManifestResourceStream("Scrabble.dictionary.dat");
            byte[]     b    = new byte[file.Length];
            
            for (int x = 0; x < b.Length; x++)
            {
                b[x] = (byte)file.ReadByte();
            }

            this._stream.Write(b, 0, b.Length);
            this._size_byte =  (int)this._stream.Length;
            this._size_bit  = (long)this._size_byte << 3;
            this.Seek(0, SeekOrigin.Begin);
            
            b = null;
            GC.Collect();
        }
        public void     SaveToFile   (string path)
        {
            if (File.Exists(path))
            {
                File.Delete(path);
            }

            FileStream file = new FileStream(path, FileMode.CreateNew);

            this._stream.WriteTo(file);

            file.Flush();
            file.Close();
        }
    }
}
